/*	Copyright (C) 2018-2024 Martin Guy <martinwguy@gmail.com>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 3
 *	as published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Program: spettro
 *	Play an audio file displaying a scrolling log-frequency spectrogram.
 *
 * File: main.c
 *	Main routine, parameter handling, reading and servicing key strokes,
 *	scheduling FFT calculations and receiving the results,
 *	the FFT result cache and drawing overlays on rows and columns.
 *
 * The audio file is given as a command-line argument
 * A window opens showing a graphical representation of the audio file:
 * each frame of audio samples is shown as a vertical bar whose colors are
 * taken from the "heat maps" of "sox spectrogram".
 *
 * The color at each point represents the energy in the sound at some
 * frequency (band) at a certain moment in time (over a short period).
 * The vertical axis, representing frequency, is logarithmic, giving an
 * equal number of pixel rows in each octave of the scale, by default
 * 9 octaves from 27.5 Hz (very bottom A) to 14080 Hz (the toppest A
 * we can hear.)
 *
 * At startup, the start of the piece is at the centre of the window and
 * the first seconds of the audio file are shown on the right half. The
 * left half is all grey.
 *
 * If you hit play (press 'space'), the audio starts playing and the
 * display scrolls left so that the current playing position remains at the
 * centre of the window. Another space pauses the playback, another makes it
 * continue from where it was. When it reaches the end of piece, the playback
 * stops; pressing space makes it start again from the beginning.
 *
 * If you resize the window the displayed image is zoomed.
 *
 * For command-line options and key bindings, see Usage in args.c.
 *
 * It runs in three types of thread:
 * - the main thread handles GUI events, starts/stops the audio player,
 *   tells the calc thread what to calculate, receives results, and
 *   displays them.
 * - The calc thread performs FFTs and reports back when they're done.
 * - the timer thread is called periodically to scroll the display in sync
 *   with the audio playback. The actual scrolling is done in the main loop
 *   in response to an event posted by the timer thread.
 *
 *	Martin Guy <martinwguy@gmail.com>, Dec 2016 - May 2017.
 */

#include "spettro.h"

/*
 * System header files
 */
#include <unistd.h>	/* for sleep() */
#include <time.h>	/* for time() */
#if SDL_AUDIO && HAVE_LIBX11
#include <X11/Xlib.h>	/* for XInitThreads() */
#endif

/*
 * Local header files
 */
#include "args.h"
#include "audio.h"
#include "audio_cache.h"
#include "axes.h"
#include "caca_text.h"
#include "result_cache.h"
#include "convert.h"
#include "gui.h"
#include "interpolate.h"
#include "overlay.h"
#include "paint.h"
#include "scheduler.h"
#include "text.h"
#include "timer.h"
#include "window.h"	/* for free_windows() */
#include "ui.h"
#include "ui_funcs.h"	/* for set_time_zoom() */

static char *progname;	/* executable name for stderr or re-exec */

int
main(int argc, char **argv)
{
    audio_file_t *af = NULL;
    char *filename;
    char **original_argv = argv;
    int optind;

    progname = argv[0];

    optind = process_args(argc, argv);
    if (optind == -1) exit(1);

    /* Set variables with derived values */
    disp_offset = disp_width / 2;
    min_x = 0; max_x = disp_width - 1;
    min_y = 0; max_y = disp_height - 1;

#if HAVE_LIBCACA
    /* Initialize the caca font here so that text size measurements are OK */
    caca_init();
#endif

    /* Width of frequency axis on the left of the graph.
     * 22050- == pixel space + text + pixel space + 2 pixels for tick.
     * With the caca font, "2E+05" is one pixel wider than "00000" but
     * it just uses the single-pixel gap on the left and is uncommon.
     */
    frequency_axis_width = 1 + text_width("00000") + 1 + 2;

    /* Note name frequency axis on the right.
     * -A0 == Two pixels for tick, a space, two * (letter + blank column)
     */
    note_name_axis_width = 2 + 1 + text_width("A0") + 1;

    /* Top and bottom axes = space above, text, space below. */
    top_margin = 1 + text_height("A") + 1;
    bottom_margin = 1 + text_height("0") + 1;

    if (show_frequency_axes) {
	min_x += frequency_axis_width;
	max_x -= note_name_axis_width;
    }
    if (show_time_axes) {
	min_y += bottom_margin;
	max_y -= top_margin;
    }

    /* Process the filename argument */
    if (optind == argc) {
	fprintf(stderr, "You must name some audio files.\n");
	exit(1);
    }
    if (shuffle) {
	/* Pick a random file */
	srand((unsigned)time(NULL));
	optind += rand() % (argc - optind);
    }
    filename = argv[optind];

    if ((af = open_audio_file(filename)) == NULL) {
	fprintf(stderr, "Cannot read ");
	perror(filename);
	goto next_file;
    }

/* If anything uses SDL, initialise it. The audio and video will be
 * enabled separately in audio_init() and gui_init()
 */
#if SDL_AUDIO || SDL_TIMER || SDL_VIDEO || SDL_MAIN

# if HAVE_LIBX11
    /* Without this, you get:
     * [xcb] Unknown request in queue while dequeuing
     * [xcb] Most likely this is a multi-threaded client and XInitThreads has not been called
     * [xcb] Aborting, sorry about that.
     */
    if (!XInitThreads()) {
	fprintf(stderr, "XInitThreads failed.\n");
	/* May be harmless */
    }
# endif

    {
	if (SDL_Init(
# if SDL_TIMER
		     SDL_INIT_TIMER |
# endif
# if SDL_MAIN
		     SDL_INIT_EVENTS |
# endif
		    0) != 0) {
	    fprintf(stderr, "Couldn't initialize SDL: %s.\n", SDL_GetError());
	    exit(1);
	}
	atexit(SDL_Quit);
    }
#endif

    /* Initialize the graphics subsystem */
    gui_init(filename);

    /* Initialize the audio subsystem */
    if (!output_file && !mute) audio_init(af, filename);

    /* If they set disp_time with -t or --start, check that it's
     * within the audio and make it coincide with the start of a column.
     */
    if (start_time >= audio_file_length()) {
	fprintf(stderr,
		"Starting time (%s) ", seconds_to_string(disp_time));
	fprintf(stderr, "is beyond the end of the audio (%s).\n",
		seconds_to_string(audio_file_length()));
	set_disp_time(audio_file_length() - 1);
    } else {
	set_disp_time(start_time);
    }

    /* The row overlay (piano notes/staff lines) doesn't depend on
     * the sample rate, only on min/max_freq, so it doesn't change
     * from file to file */
    make_row_overlay();	

    /* Apply the -t flag */
    if (disp_time != 0.0) set_playing_time(disp_time);
 
    start_scheduler(max_threads);

    draw_axes();

    repaint_display(FALSE); /* Schedules the initial screen refresh */

    /* Start the screen-scrolling interrupt */
    if (!output_file) start_timer();

    /* Run the main event loop until we quit */
    gui_main();

    if (!output_file) stop_timer();
    stop_scheduler();
    if (!output_file && !mute) audio_deinit();
    gui_deinit();

next_file:
    if (nexting && optind < argc-1) {
	unsigned i;
	/* Shift the following filenames down, removing the current filename. */
	for (i=optind; i<argc-1; i++) {
	    argv[i] = argv[i+1];
	};
	/* Make sure argv[] is NULL-terminated for execvp() */
	argv[i] = NULL;

	execvp(progname, original_argv);
	fprintf(stderr, "Re-exec of %s failed.\n", progname);
	exit(1);
    }

    /* Free memory to make valgrind happier */
    drop_all_work();
    drop_all_results();
    free_interpolate_cache();
    free_row_overlay();
    free_windows();
    close_audio_file(af);

    return 0;
}
